#ifndef ASSIGNINFOS_HPP_
#define ASSIGNINFOS_HPP_
#define assigninfos_hpp_version      "$Revision: 1.15 $"
#include "BitVector.hpp"
#include "Types.hpp"
#include "WatchdCls.hpp"


class VarVal{
	TwoBitsVec<Variable> vals;
public:
	VarVal():vals(){}
	VarVal(const VarVal& vls,const bool deep):vals(vls.vals,deep){}
	uint32_t 		count() const{return vals.getSize();}
	/*bool			takeThisIf(const VarVal& v){
					if(v.vals.getSize() == vals.getSize()){ return false;}
					vals = v.vals; return true;}*/
	void			ensureVarVals(uint32_t sz){
					asrt(sz > vals.getSize());
					vals.ensureSize(sz,true);}

	/*void			initWith(const VarVal& from){
					vals = from.vals;}*/


	bool			varHasValue(const Variable& v, bool b)const{
					return vals.hasVal(v,2+(uint32_t)b);}

	bool			varPrevValWasTrue(const Variable& v)const {
					return vals.secSet(v);}

	void			litSetPrevVal(const Literal& l){
					l.isPos() ? vals.setSec(l.toVar()) : vals.clearSec(l.toVar());}

	bool			litPrevValWas(const Literal& l)const {
					return vals.secSet(l.toVar()) == l.isPos();}

	void			varClearValue(const Variable& v){
					vals.clearFirst(v);}

	void			varClearValueFlipPrev(const Variable& v){
					vals.setVal(v,!vals.secSet(v));}

	bool			varKnwn(const Variable& v)const{
					return vals.firstSet(v);}

		//Literals
	bool			litTrue(const Literal& l)const 	{return varHasValue(l.toVar(),l.isPos());}
	bool			litFalse(const Literal& l)const	{return varHasValue(l.toVar(),l.isNeg());}
	bool			litUkwn(const Literal& l)const	{return !varKnwn(l.toVar());}
	bool			litKnwn(const Literal& l)const	{return varKnwn(l.toVar());}

	void			setValue(const Variable& var, const bool& pos){
					vals				.setVal(var,2+((uint32_t)pos));
					asrt(varKnwn(var));	}

	bool			setValueChng(const Variable& var, const bool& pos){
					uint64_t prev = vals.setGetVal(var,2+((uint32_t)pos));asrt(varKnwn(var)); asrt(prev < 2);
					return prev != pos;}

	void			setValue(const Literal& lit){
					setValue(lit.toVar(), lit.isPos());}

};

const uint32_t MAX_R_VAL		= max32 >> 1;
const uint32_t CORE_DONT_CARE	= 0;
const uint32_t CORE_TEST_CAND	= MAX_R_VAL;

class Reason{
	static const uint32_t DEC 		= 0;
	static const uint32_t CLS 		= 1;
	static const uint32_t UNIT		= 2; //only for real unit clauses without reason (in input formula)
	static const uint32_t CSTAT_FIX	= 3; //only for real unit clauses without reason (in input formula)



	uint64_t	reason	: 	31; //clp

	//state the kind of reason
	uint64_t	type 	: 	2; //Decision / Cls / UNIT

	//cache lbd in reasons to avoid to touch the clauses
	uint64_t	lbd		:	4;

	//decision level
	uint64_t	level	:	27;





	//uint64_t	tern_reas: 	31; //reason for special ternaries


public:
	static const uint32_t LBD_MAX 		= 16;
	static const uint32_t LBD_RND 		= 10;

	//static const uint32_t CORE_TEST		= NEVER_MCORE-1;


	Reason():reason(MAX_R_VAL),type(DEC),lbd(LBD_MAX-1),level(max32>>5){}

	//void		init(){reason=max32>>1;type=DEC; lbd = LBD_MAX-1;level = max32>>5;}
	//bool		isShrtReason()		const	{ return type == SHRT;}
	bool		isClsReason()		const	{ return type == CLS;}
	bool		isDecision()		const	{ return type == DEC;}
	bool		isUnit()			const	{ return type >= UNIT;}
	bool		isUnitFix()			const	{ return type == CSTAT_FIX;}

	//bool		isReason(const ClsP cls)const
	//			{return isClsReason() && cls== Ptr32(reason);}

	//Literal		getShrtFrst()		const	{ asrt(isShrtReason());return Literal(reason);}
	//Literal		getShrtSec()		const	{ asrt(isShrtReason());return Literal(tern_reas);}

	ClsP		getClsReason()  	const	{ asrt(isClsReason());return ClsP(reason);}
	uint32_t	getUnitClsID()		const	{ asrt(isUnit()); return reason;}
	uint32_t	getLevel()			const	{ return level;}
	uint32_t	getLBD()			const	{ asrt(!isUnit()); return lbd+1;}
	bool		didQuickTest()		const	{ asrt(isUnit()); return lbd == 0;}
	bool		setCoreTestNew()			{ asrt(isUnit()); bool ret = lbd; lbd=0; return ret;}



	/*void 		setShrtReason(const uint32_t lev, const Literal& a, const Literal& b){
				asrt(level!=0); asrt(type != UNIT);
				lbd=LBD_MAX-1; level=lev; type = SHRT; reason= a.idx(); tern_reas = b.idx();}*/

	void 		setClsReason(const uint32_t lev, const ClsP cls, const uint32_t LBD){
				asrt(level!=0); asrt(LBD>0); asrt(LBD<=LBD_MAX); asrt(!isUnit());
				reason 	= cls.idx(); 
 				type 	= CLS; 
				lbd		= LBD-1; 
				level	= lev;}

	void 		setDecision(const uint32_t lev){
				asrt(level!=0); asrt(lev!=0); asrt(!isUnit());
				type 	= DEC;
				lbd		= LBD_MAX-1;
				level	= lev;}

	void		setIniUnit(const uint32_t cls_id,const bool stat_fix){asrt(type != CSTAT_FIX);
				reason 	= cls_id;
				type 	= UNIT + stat_fix;
				//DONT SET LBD SINCE ITS USED TO MARK UNITS FOR QUICK SAT CHECKS IN CORE MINIMISATION
				level	= 0;
				asrt(stat_fix || type==UNIT); asrt(!stat_fix || type==CSTAT_FIX);}

	//void		setInMinCore(){asrt(level==0);asrt(type==UNIT);
				//type=MCORE;}

	void		setLevel(const uint32_t& l){asrt(l != 0); asrt(type != CSTAT_FIX); level = l; asrt(level==l);}
	void		initLevel(){setLevel(max32>>5); }

	bool		isFix()const {return level==0;}
};



//template<class Reason>
class AssReason{
	AssReason(const AssReason& v){asrt(false);}
	AssReason& operator=(const AssReason& v){asrt(false);return *this;}
protected:
	einfo<Reason,Variable>	reasons;
public:
	AssReason(){};
	uint32_t varsCount() const {return reasons.size();}

	void	ensureVars(uint32_t sz){
			asrt(sz > reasons.size());
			reasons		.ensureSize(sz);}

	uint32_t varAssLev(const Variable& v) const {
			return reasons[v].getLevel();}

	uint32_t varUnitClsID(const Variable& v)const{ asrt(reasons[v].isUnit());
			return reasons[v].getUnitClsID();}

	bool	varIsUnit(const Variable& v)const{
			return reasons[v].isUnit();}

	bool	varUnitFix(const Variable& v)const{
			return reasons[v].isUnitFix();}

	bool 	varFix(const Variable& v) const {
			return reasons[v].isFix();}

	bool	varIsDecision(const Variable& v) const {
			return reasons[v].isDecision();}

	bool	setCoreTestNew(const Literal& lit) {
			return reasons[lit.toVar()].setCoreTestNew();}

	bool	didQuickTest(const Variable& v) const{
			return reasons[v].didQuickTest();}

	/*bool	varHasReason(const Variable& v,const ClsP cls) const{
			return reasons[v].isReason(cls);}*/

	/*bool	varIsShrtReason(const Variable& v) const{
			return reasons[v].isShrtReason();}*/

	const 	Reason& varAssignReason(const Variable& v) const{
			return reasons[v];}

	/*Literal	varGetShrtFrstReason(const Variable& v) const{
			return reasons[v].getShrtFrst();}

	Literal	varGetShrtSecReason(const Variable& v) const{
			return reasons[v].getShrtSec();}*/

	void	litSetLev(const Literal& l, uint32_t levl){ asrt(levl != 0);
			reasons[l.toVar()].setLevel(levl);}

	//Literals
	const uint32_t		litAssLev(const Literal& l)		const 	{return varAssLev(l.toVar());}

	bool 				litFix(const Literal& l) 		const {return varFix(l.toVar());}
};







//template<class Reason>
class Assigns : public AssReason{
	VarVal&		vals;
	var_hp_t& 	act_hp;
	uint32_t	agility;
	const bool 	bump_agil;

	void setVal(const Variable& v, const bool p){
		agility -= agility/10000;
		if(vals.setValueChng(v,p)){
			agility += 1000;
			if(bump_agil){act_hp.activityBump(v);}
		}
	}

public:
	bool lessAgilThan(const uint32_t& x) const {return agility < x;}
	Assigns(VarVal& vls,var_hp_t& v_act_hp,const bool bump)
		:AssReason(),vals(vls),act_hp(v_act_hp),agility(0),bump_agil(bump){};
	VarVal& getValsRef() {return vals;}
	uint32_t reasonLBD(const Variable& var) const{
			asrt(reasons.inRange(var)); asrt(vals.varKnwn(var));
			return reasons[var].getLBD();}

	void	setValIniUnit(const Literal& lit, uint32_t cls_id,const bool stat_fix){
			const Variable var 	= lit.toVar(); 	asrt(reasons.inRange(var));
			vals.setValue(var,lit.isPos());		asrt(vals.varKnwn(lit.toVar()));asrt(vals.varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setIniUnit(cls_id,stat_fix);}

	//void	putInMinCore(const Variable& var){
		//	reasons[var].setInMinCore();}

	//void	flipUnitValTo(const Literal& lit){ asrt(vals.litFalse(lit));
			//vals.setValue(lit); asrt(reasons[lit.toVar()].isUnit());}


	void	setValueDecision(const Literal& lit, uint32_t lev){ asrt(lev != 0);
			const Variable var 	= lit.toVar(); 	asrt(reasons.inRange(var));
			vals.setValue(var,lit.isPos());		asrt(vals.varKnwn(var));asrt(vals.varPrevValWasTrue(var)==lit.isPos());
			reasons[var]		.setDecision(lev);}


	void	setValueReason(const Literal& lit, uint32_t lev,const ClsP cls, const uint32_t lbd){
			const Variable var 	= lit.toVar(); 	asrt(reasons.inRange(var)); asrt(!reasons[var].isUnit() || (vals.varPrevValWasTrue(var)==lit.isPos()));
			vals.setValue(var,lit.isPos());		asrt(vals.varKnwn(var));asrt(vals.varPrevValWasTrue(var)==lit.isPos());
			reasons[var]		.setClsReason(lev,cls,lbd);}

	void	clearL0Value(const Variable& var){
			vals			.varClearValue(var);
			reasons[var]	.initLevel();}


	/*void	setValueReason(const Literal& lit, uint32_t lev,const Literal& r1, const Literal& r2){
			const Variable var 	= lit.toVar(); 	asrt(reasons.inRange(var));
			vals.setValue(var,lit.isPos());		asrt(vals.varKnwn(lit.toVar()));asrt(vals.varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setShrtReason(lev,r1,r2);}*/

	bool 	litFixTrue(const Literal& l) 	const {return vals.litTrue(l) && varFix(l.toVar());}
	bool 	litFixFalse(const Literal& l) 	const {return vals.litFalse(l) && varFix(l.toVar());}
};




#endif /* WATCHANDREASON_HPP_ */
